-- | A simple abstract type to help create well-formed HTML documents
--
-- * Ensure that entities are not escaped twice, e.g  @"\<"@ ⟹ @"&lt;"@ ⟹ @"&amp;lt;"@
-- * Ensure that attribute values are quoted when needed
-- * Ensure that tags are properly nested

module HTML(
    HTML,URL,Tag,   
    (<>),mempty,
    -- ** Create a complete HTML page
    page,
    -- ** HTML5 body parts
    header,main_,footer,
    -- ** Create HTML elements and text
    h1,h2,h3,
    ul,ol,li,p,
    link,optlink,text,
    -- ** Forms
    form,input,submit,
    -- ** Insert arbitrary markup, no guarantees
    markup) where
import Data.Monoid(Monoid(..))
import Data.Semigroup
import Data.Char(isAlphaNum)

type URL = String
type Tag = String

newtype HTML = HTML String
instance Show HTML where showsPrec _ (HTML s) = showString s

instance Semigroup HTML where
  HTML s1 <> HTML s2 = HTML (s1++s2)
  
instance Monoid HTML where
  mempty = HTML ""
  mappend = (<>)

h1 = el "h1" . text
h2 = el "h2" . text
h3 = el "h3" . text

header = el "header"
footer = el "footer"
main_ = el "main"

ul = el "ul"
ol = el "ol"
li item = tag "li"<> item

p e = tag "p" <> e

link :: URL -> HTML -> HTML
link url = el' "a" [("href",url)]
optlink = maybe id link


el t s = tag t<>s<>endtag t

el' :: Tag -> [(String,String)] -> HTML -> HTML
el' t as s = tag' t as<>s<>endtag t

tag :: Tag -> HTML
tag t = HTML ("<"++t++">")
endtag t = tag ('/':t)

tag' :: Tag -> [(String,String)] -> HTML
tag' t as = tag (t++shas as)
  where
    shas = concatMap sha
    sha (n,v) = ' ':n++"="++quote (esc v)

esc :: String->String
esc = concatMap esc1
  where esc1 '<' = "&lt;"
        esc1 '>' = "&gt;"
	esc1 '&' = "&amp;"
	esc1 c = [c]

quote s =
  if all isAlphaNum s
  then s
  else '"':s++"\""

text = HTML . esc
markup = HTML -- !!!

page :: String -> URL -> HTML -> HTML
page title style body =
    markup "<!DOCTYPE html>" <>
    el "html" (
    el "head" (el "title" (text title)<>
               tag' "meta" [("charset","UTF-8")] <>
               tag' "link" [("rel","stylesheet"),("href",style)]) <>
    el "body" body)


form :: URL -> HTML -> HTML
form action html = el' "form" [("action",action)] html

input name ph = tag' "input" [("name",name),("placeholder",ph)]
submit label = tag' "input" [("type","submit"),("value",label)]
