-- | MimiIMDB - a web server that implements a mini version of www.imdb.com
import Network.Shed.Httpd
import Database.HDBC
import Database.HDBC.PostgreSQL
import Network.URI
import Codec.Binary.UTF8.String(encodeString)
import HTML
import SqlRows

main = do db <- connectPostgreSQL ""
          initServer 8080 (miniImdb db)

-- | HTTP Request handler
miniImdb db request@Request{reqMethod="GET",reqURI=url} = miniImdbGet db url
miniImdb db req = return badRequestResponse

-- | HTTP GET request handler
miniImdbGet db url =
  case (uriPath url,inputs (uriQuery url)) of
    ("/",[]) -> return (htmlResponse startPage)
    ("/title",[("tt",tt)]) -> miniImdbShowTitle db tt
    ("/name", [("nm",nm)]) -> miniImdbShowName db nm
    ("/title",[("q",q)]) -> miniImdbSearchTitle db q
    ("/name", [("q",q)]) -> miniImdbSearchName db q
    ("/miniimdb.css", _) -> okResponse "text/css" <$> readFile "miniimdb.css"
    _ -> return badRequestResponse
  where
    inputs = queryToArguments . fixplus
      where
        fixplus = concatMap decode
        decode '+' = "%20" -- httpd-shed bug workaround
        decode c   = [c]

--------------------------------------------------------------------------------
-- ** Search result pages

-- | Title search result page
miniImdbSearchTitle db q = titleList <$> findTitles db q
  where
    titleList titles = 
       htmlResponse $
       page2 ("Title search result for "++q) $
             ol (mconcat (map (li.titleLink) titles))

-- | Name search result page
miniImdbSearchName db q = nameList <$> findNames db q
  where
    nameList names = 
      htmlResponse $
      page2 ("Name search result for "++q) $
            ol (mconcat (map (li.nameLink) names))

-- | Name info page
miniImdbShowName db nm = person <$> nameInfo db nm <*> nameTitles db nm
  where
    person info@(name,y1,y2,pp) titles =
      htmlResponse $
      page2 (name++optYears y1 y2) $
      p (maybe mempty text pp) <>
      p (link ("https://www.imdb.com/name/"++nm) (text "IDMB"))<>
      h3 "Filmography" <>
      ol (mconcat (map (li.titleInfo) titles))

    titleInfo (tt,title,y1,y2,cat,cs,r) =
      titleLink (tt,title,y1,y2)<>optParen r<>
      text " "<>maybe mempty text cat<>
      text " "<>maybe mempty text cs

-- | Title info page
miniImdbShowTitle db tt = title <$> titleInfo db tt <*> titleNames db tt
  where
    title info@(title,y1,y2,genres) names =
      htmlResponse $
      page2 (title++optYears y1 y2) $
      p (maybe mempty text genres)<>
      p (link ("https://www.imdb.com/title/"++tt)(text "IDMB"))<>
      h3 "Principal cast & crew " <>
      ol (mconcat (map (li.nameInfo) names))

    nameInfo (nm,name,y1,y2,cat,cs) =
      nameLink (nm,name,y1,y2)<>
      text " "<>maybe mempty text cat<>
      text " "<>maybe mempty text cs

--------------------------------------------------------------------------------
-- ** HTML output helper functions

titleLink :: (TConst,Title,Year,Year) -> HTML
titleLink (tt,title,y1,y2) =
    link ("/title?tt="++tt) (text title)<>text (optYears y1 y2)

nameLink :: (NConst,Name,Year,Year) -> HTML
nameLink  (nm,name,y1,y2)  =
    link ("/name?nm="++nm) (text name)<>text (optYears y1 y2)

optParen x = maybe mempty (parens.text.show) x
parens s = text " ("<>s<>text ")"

optYears :: Year -> Year -> String
optYears y1 y2 =
  case (y1,y2) of
    (Just y1,Just y2) -> " ("++show y1++"-"++show y2++")"
    (Just y1,Nothing) -> " ("++show y1++")"
    (Nothing,Just y2) -> " (?-"++show y2++")"
    (Nothing,Nothing) -> ""

startPage = page2 "Start page" body
  where
    body = h3 "Title search" <> search "title"
           <>
           h3 "Name search" <> search "name"

    search name = form name (input "q" name <> submit "Search")

page2 subtitle body =
  show $
  page (title++" - "++subtitle) "/miniimdb.css"
       (header (h1 title) <>
        main_ (h2 subtitle<>body)<>
        footer (link "/" (text "MiniIMDB")))
  where
    title = "MiniIMDB"

---------------------------------------------------------------------------------- ** HTTP responses

badRequestResponse = Response 400 [] "Bad request"

okResponse ty = Response 200 [("Content-Type",ty)] . encodeString

textResponse = okResponse "text/plain"
htmlResponse = okResponse "text/html"

--------------------------------------------------------------------------------
-- ** Database access

type NConst = String
type TConst = String
type Name = String
type Title = String
type Year = Maybe Int
type Professions = Maybe String
type Category = Maybe String
type Characters = Maybe String
type Genres = Maybe String
type Rating = Maybe Double

findNames :: Connection -> Name -> IO [(NConst,Name,Year,Year)]
findNames db q =
    fromRowsM =<< quickQuery' db query [toSql q]
  where
    query = "SELECT nconst,primaryName,birthYear,deathYear "++
            "FROM Name_basics WHERE primaryName LIKE ?"

findTitles :: Connection -> Title -> IO [(TConst,Title,Year,Year)]
findTitles db q =
    fromRowsM =<< quickQuery' db query [toSql q]
  where
    query = "SELECT tconst,originalTitle,startYear,endYear "++
            "FROM Title_basics WHERE originalTitle LIKE ?"
          ++" AND titletype IN ('movie','tvSeries')"
          ++" ORDER BY startYear"

nameInfo :: Connection -> NConst -> IO (Name,Year,Year,Professions)
nameInfo db nm = oneRowM =<< quickQuery' db query [toSql nm]
  where
    query = "SELECT primaryName,birthYear,deathYear,primaryProfession "
          ++"FROM Name_Basics WHERE nconst = ?"

titleInfo :: Connection -> TConst -> IO (Title,Year,Year,Genres)
titleInfo db tt =
    oneRowM =<< quickQuery' db query [toSql tt]
  where
    query = "SELECT originalTitle,startYear,endYear,genres "++
            "FROM Title_basics WHERE tconst = ?"

nameTitles :: Connection -> NConst ->
              IO [(TConst,Title,Year,Year,Category,Characters,Rating)]
nameTitles db nm =
    fromRowsM =<< quickQuery' db query [toSql nm]
  where
    query = "SELECT tconst,originalTitle,startYear,endYear,category,characters,averageRating "
          ++"FROM Titles_names_ratings "
          ++"WHERE nconst =? AND titleType in ('movie','tvSeries') "
          ++"ORDER BY startYear"


titleNames :: Connection -> TConst ->
              IO [(NConst,Name,Year,Year,Category,Characters)]
titleNames db tt =
    fromRowsM =<< quickQuery' db query [toSql tt]
  where
    query = "SELECT nconst,primaryName,birthYear,deathYear,category,characters "
          ++"FROM Titles_names_ratings "
          ++"WHERE tconst = ? AND titleType in ('movie','tvSeries') "

oneRowM [row] = fromRowM row
oneRowM _ = fail "expected one row"
