-- | Student Portal User Interface with Threepenny GUI
module StudentPortal3p(studentPortal) where
import System.IO.Error(tryIOError)
import Graphics.UI.Threepenny.Core as UI
import qualified Graphics.UI.Threepenny as UI

import StudentPortalCommon

studentPortal :: StudentPortalFunctions -> IO ()
studentPortal dbi = startGUI defaultConfig (studentPortal3p dbi)

studentPortal3p dbi window = startPage
  where
    startPage =
      do main <- mkElement "main"
         getHead window #+ [mkElement "style" #+ [string stylesheet]]
         getBody window #+ [UI.h1 #+ [string "Student Portal"],pure main]
         loginPage main

    loginPage main =
      do lbl <- mkHTML "Student idnr: "
         user <- mkInput 20 "" # set (attr "placeholder") "idnr"
         btn  <- mkButton "Login"
         emsg <- UI.span
         pure main # set UI.children [lbl,user,btn,emsg]
         let login = loginAs main emsg =<< get value user
         on UI.click btn $ \ _ -> login
         on valueChange' user $ \ _ -> login

    loginAs main emsg idnr =
      do r <- tryIO (getInformation dbi idnr)
         case r of
           Left e -> do pure emsg # set children []
                        pure emsg #+ [string (show e)]
                        return ()
           Right info -> loggedIn main emsg idnr info

    loggedIn main emsg idnr info =
      do pure main # set UI.children []
         [logout,reload,reg,unreg] <-
             mapM mkButton ["Logout","Refresh","Register","Unregister"]
         lbl <- mkHTML "      Course: "
         code <- mkInput 10 "" # set (attr "placeholder") "code"
         info <- mkHTML (showInformation info)
         pure main #+ map pure [logout,reload,lbl,code,reg,unreg,emsg,info]
                       
         on UI.click logout $ \ _ -> loginPage main
         on UI.click reload $ \ _ -> loginAs main emsg idnr
         on UI.click reg $ \ _ -> register =<< get value code
         on UI.click unreg $ \ _ -> unregister =<< get value code
         return ()
       where
         register code =
           showResult =<< tryIO (registerStudent dbi idnr code)
         unregister code =
           showResult =<< tryIO (unregisterStudent dbi idnr code)

         showResult r =
           do msg <- string (either show id r)
              pure emsg # set children [msg]

stylesheet =
  unlines [
    "body { background: #eee; }",
    "h1 { border-bottom: 2px solid #006; }",
    "h1,h2,caption,th { color: #006; font-family: sans-serif; }",
    "table { display: inline-table; vertical-align: top; margin: 2ex;",
    "        border-collapse: collapse; }",
    "th { background: #ffe; }",
    "caption { font-size: 110%; white-space: nowrap; }",
    "th,td { padding: 4px; border: 1px solid #666; }",
    "th { text-align: left; }",
    "td { background: white;"]

--------------------------------------------------------------------------------

showInformation :: StudentInformation -> String
showInformation (basic,readcs,regcs,unreadcs,pathToGrad) =
   unlines [showBasicInformation basic,
            table "Read courses" (
              (tr++th++"Name (code)"++th++"Credits"++th++"Grade"):
              map showReadCourse readcs),
            table "Registered courses" (
              (tr++th++"Name (code)"++th++"Status"):
              map showRegisteredCourse regcs),
            table "Unread mandatory courses" (
              (tr++th++"Name (code)"):
              map showUnreadCourse unreadcs),
            table "Path to graduation" (showPathToGraduation pathToGrad)]
  where
    showBasicInformation (idnr,name,login,program,branch) =
      unlines [h2 ("Information for student "++idnr),
               table "Student" [
                 tr++th++"Name: "++td ++ name,
                 tr++th++"Student ID: "++td ++ login,
                 tr++th++"Program: "++td ++ program,
                 tr++th++"Branch: "++td ++ maybe "-" id branch]]

    showReadCourse (code,name,credits,grade) =
      tr++td++name++" ("++code++")"++td++show credits++"p"++td++grade

    showRegisteredCourse (code,name,status) =
      tr++td++name++" ("++code++")"++td++showStatus status

    showUnreadCourse (name,code) = tr++td++name++" ("++code++")"

    showPathToGraduation (PtG tc mc rc sc ml q) =
      [tr++th++"Seminar courses taken: "++td++show sc,
       tr++th++"Math credits taken: "++td++show mc++"p",
       tr++th++"Research credits taken: "++td++show rc++"p",
       tr++th++"Total credits taken: "++td++show tc++"p",
       tr++th++"Mandatory courses left: "++td++show ml,
       tr++th++"Fullfills the requirements for graduation: "++td++
                 (if q then "yes" else "no")]

    showStatus Registered = "registered"
    showStatus (Waiting n) = "waiting as nr "++show n

--------------------------------------------------------------------------------
-- * Utilties

tryIO io = liftIO (tryIOError io)

--------------------------------------------------------------------------------
-- ** Simple HTML generation

table caption rows =
  unlines (start "table":
           (el "caption" caption):
           rows++
           [end "table"])

h2 = el "h2"
tr = start "tr"
th = start "th"
td = start "td"

el tag txt = start tag ++ txt ++ end tag
start tag = '<':tag++">"
end tag = start ('/':tag)

--------------------------------------------------------------------------------
-- ** Threepenny utilities

-- | @mkInput width init@ makes an input element with the specified width and
-- initial text
mkInput :: Int -> String -> UI Element
mkInput width init = do
    UI.input # set UI.type_ "text"
             # set UI.size (show width)
             # set value init

-- | @mkButton label@ makes a clickable button with the given label
mkButton :: String -> UI Element
mkButton label = UI.button #+ [string label]


-- | @mkHTML html@ makes an element with the specified HTML content
mkHTML :: String -> UI Element
mkHTML html = UI.span # set UI.html html

-- | Event that occurs when the presses enter in an input element.
valueChange' :: Element -> Event String
valueChange' el = unsafeMapUI el (const $ get value el) (domEvent "change" el)
  where
    unsafeMapUI el f = unsafeMapIO (\a -> getWindow el >>= \w -> runUI w (f a))
