-- | Student Portal Command Line Interface
module StudentPortalCLI(studentPortal) where
import System.Environment
import Data.List(isPrefixOf)
import System.IO(hFlush,stdout)
import System.IO.Error(catchIOError,ioeGetErrorString,isUserError)

import StudentPortalCommon

studentPortal :: StudentPortalFunctions -> IO ()
studentPortal dbi =
  do args <- getArgs
     idnr <- case args of
               n:_ -> return n
               []  -> while null (ask "Input student idnr: ")
     putStrLn usage
     putStrLn "Welcome!"
     studentPortalCLI dbi idnr
     return ()

-- Read-eval-print loop
studentPortalCLI conn idnr =
   while id $
     maybe oops runCommand . parseCommand . words =<< while null (ask "? > ")
  where
    oops = do putStrLn "Unrecognized command"
              output usage

    runCommand cmd =
      handleIOErrors $
      case cmd of
        Information     -> output . showInformation =<< getInformation conn idnr
        Register course   -> output =<< registerStudent conn idnr course
        Unregister course -> output =<< unregisterStudent conn idnr course
        Quit              -> return False

    output x = do putStrLn x
                  return True
--------------------------------------------------------------------------------

showInformation :: StudentInformation -> String
showInformation (basic,readcs,regcs,unreadcs,pathToGrad) =
   unlines [showBasicInformation basic,
            "Read courses (name (code), credits: grade):",
            unlines (map showReadCourse readcs),
            "Registered courses (name (code): status):",
            unlines (map showRegisteredCourse regcs),
            "Unread mandatory courses:",
            unlines (map showUnreadCourse unreadcs),
            showPathToGraduation pathToGrad]
  where
    showBasicInformation (idnr,name,login,program,branch) =
      unlines ["Information for student "++idnr,
               "----------------------------------",
               "Name      : " ++ name,
               "Student ID: " ++ login,
               "Program   : " ++ program,
               "Branch    : " ++ maybe "-" id branch]

    showReadCourse (code,name,credits,grade) =
      "  "++name++" ("++code++"), "++show credits++"p: "++grade

    showRegisteredCourse (code,name,status) =
      "  "++name++" ("++code++")"++": "++showStatus status

    showUnreadCourse (name,code) = "  "++name++" ("++code++")"

    showPathToGraduation (PtG tc mc rc sc ml q) =
      unlines ["Seminar courses taken:  "++show sc,
               "Math credits taken:     "++show mc,
               "Research credits taken: "++show rc,
               "Total credits taken:    "++show tc,
               "Mandatory courses left: "++show ml,
               "Fullfills the requirements for graduation: "++
                         (if q then "yes" else "no")]

    showStatus Registered = "registered"
    showStatus (Waiting n) = "waiting as nr "++show n

--------------------------------------------------------------------------------
-- * Command line syntax and parsing

data Command = Information
             | Register CourseCode
             | Unregister CourseCode
             | Quit

usage =
  unlines
     ["Usage:",
      "       i[nformation]",
      "       r[egister] <course>",
      "       u[nregister] <course>",
      "       q[uit]"]

parseCommand [] = Nothing
parseCommand (c:args) =
    case ([cmd|cmd<-cmds,c `isPrefixOf` cmd],args) of
      (["information"],[]   ) -> Just Information
      (["register"],   [arg]) -> Just (Register arg)
      (["unregister"], [arg]) -> Just (Unregister arg)
      (["quit"],       []   ) -> Just Quit
      _                       -> Nothing
  where
    cmds = ["information","register","unregister","quit"]

-------------------------------------------------------------------------------
-- * Generic utility functions

-- | Repeat a monadic action while it returns True
while p m = do a <- m
               if p a then while p m else return a

-- | Show a prompt and wait for input
ask p = do putStr p
           hFlush stdout
           getLine

-- | Catch errors and show them in a nice way
handleIOErrors io = catchIOError io handler
  where
    handler e = do if isUserError e
                   then putStrLn (ioeGetErrorString e)
                   else print e
                   return True
